@@ -10,21 +10,22 @@
 #include "scheduler.h"
 #include "scheduler_p.h"
 
-#include "sessiondata_p.h"
-#include "slaveconfig.h"
-#include "slave.h"
 #include "connection_p.h"
 #include "job_p.h"
+#include "sessiondata_p.h"
+#include "slave.h"
+#include "workerconfig.h"
 
-#include <kprotocolmanager.h>
 #include <kprotocolinfo.h>
-//#include <KJobWidgets>
+#include <kprotocolmanager.h>
 
+#ifndef KIO_ANDROID_STUB
+#include <QDBusConnection>
+#include <QDBusMessage>
+#endif
 #include <QHash>
 #include <QThread>
 #include <QThreadStorage>
-#include <QDBusConnection>
-#include <QDBusMessage>
 
 // Slaves may be idle for a certain time (3 minutes) before they are killed.
 static const int s_idleSlaveLifetime = 3 * 60;
@@ -46,13 +47,97 @@ static inline void startJob(SimpleJob *job, Slave *slave)
     SimpleJobPrivate::get(job)->start(slave);
 }
 
-// here be uglies
-// forward declaration to break cross-dependency of SlaveKeeper and SchedulerPrivate
-static void setupSlave(KIO::Slave *slave, const QUrl &url, const QString &protocol,
-                       const QStringList &proxyList, bool newSlave, const KIO::MetaData *config = nullptr);
-// same reason as above
-static Scheduler *scheduler();
-static Slave *heldSlaveForJob(SimpleJob *job);
+class KIO::SchedulerPrivate
+{
+public:
+    SchedulerPrivate()
+        : q(new Scheduler())
+    {
+    }
+
+    ~SchedulerPrivate()
+    {
+        removeSlaveOnHold();
+        delete q;
+        q = nullptr;
+        qDeleteAll(m_protocols); // ~ProtoQueue will kill and delete all slaves
+    }
+
+    SchedulerPrivate(const SchedulerPrivate &) = delete;
+    SchedulerPrivate &operator=(const SchedulerPrivate &) = delete;
+
+    Scheduler *q;
+
+    Slave *m_slaveOnHold = nullptr;
+    QUrl m_urlOnHold;
+    bool m_ignoreConfigReparse = false;
+
+    SessionData sessionData;
+
+    void doJob(SimpleJob *job);
+#if KIOCORE_BUILD_DEPRECATED_SINCE(4, 5)
+    void scheduleJob(SimpleJob *job);
+#endif
+    void setJobPriority(SimpleJob *job, int priority);
+    void cancelJob(SimpleJob *job);
+    void jobFinished(KIO::SimpleJob *job, KIO::Slave *slave);
+    void putSlaveOnHold(KIO::SimpleJob *job, const QUrl &url);
+    void removeSlaveOnHold();
+    Slave *getConnectedSlave(const QUrl &url, const KIO::MetaData &metaData);
+    bool assignJobToSlave(KIO::Slave *slave, KIO::SimpleJob *job);
+    bool disconnectSlave(KIO::Slave *slave);
+    Slave *heldSlaveForJob(KIO::SimpleJob *job);
+    bool isSlaveOnHoldFor(const QUrl &url);
+    void updateInternalMetaData(SimpleJob *job);
+
+    MetaData metaDataFor(const QString &protocol, const QStringList &proxyList, const QUrl &url);
+    void
+    setupSlave(KIO::Slave *slave, const QUrl &url, const QString &protocol, const QStringList &proxyList, bool newSlave, const KIO::MetaData *config = nullptr);
+
+    void slotSlaveDied(KIO::Slave *slave);
+
+#ifndef KIO_ANDROID_STUB
+    void slotReparseSlaveConfiguration(const QString &, const QDBusMessage &);
+#endif
+    void slotSlaveOnHoldListChanged();
+
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 91)
+    void slotSlaveConnected();
+    void slotSlaveError(int error, const QString &errorMsg);
+#endif
+
+    ProtoQueue *protoQ(const QString &protocol, const QString &host);
+
+private:
+    QHash<QString, ProtoQueue *> m_protocols;
+};
+
+static QThreadStorage<SchedulerPrivate *> s_storage;
+static SchedulerPrivate *schedulerPrivate()
+{
+    if (!s_storage.hasLocalData()) {
+        s_storage.setLocalData(new SchedulerPrivate);
+    }
+    return s_storage.localData();
+}
+
+Scheduler *Scheduler::self()
+{
+    return schedulerPrivate()->q;
+}
+
+SchedulerPrivate *Scheduler::d_func()
+{
+    return schedulerPrivate();
+}
+
+// static
+Scheduler *scheduler()
+{
+    return schedulerPrivate()->q;
+}
+
+////////////////////////////
 
 int SerialPicker::changedPrioritySerial(int oldSerial, int newPriority) const
 {
@@ -83,7 +168,7 @@ void SlaveKeeper::returnSlave(Slave *slave)
 
 Slave *SlaveKeeper::takeSlaveForJob(SimpleJob *job)
 {
-    Slave *slave = heldSlaveForJob(job);
+    Slave *slave = schedulerPrivate()->heldSlaveForJob(job);
     if (slave) {
         return slave;
     }
@@ -132,7 +217,7 @@ void SlaveKeeper::scheduleGrimReaper()
     }
 }
 
-//private slot
+// private slot
 void SlaveKeeper::grimReaper()
 {
     QMultiHash<QString, Slave *>::Iterator it = m_idleSlaves.begin();
@@ -141,11 +226,10 @@ void SlaveKeeper::grimReaper()
         if (slave->idleTime() >= s_idleSlaveLifetime) {
             it = m_idleSlaves.erase(it);
             if (slave->job()) {
-                //qDebug() << "Idle slave" << slave << "still has job" << slave->job();
+                // qDebug() << "Idle slave" << slave << "still has job" << slave->job();
             }
-            slave->kill();
             // avoid invoking slotSlaveDied() because its cleanup services are not needed
-            slave->deref();
+            slave->kill();
         } else {
             ++it;
         }
@@ -286,7 +370,7 @@ bool ConnectedSlaveQueue::removeSlave(Slave *slave)
     }
     PerSlaveQueue &jobs = it.value();
     // we need a copy as kill() ends up removing the job from waitingList
-    const QList <SimpleJob *> waitingJobs = jobs.waitingList;
+    const QList<SimpleJob *> waitingJobs = jobs.waitingList;
     for (SimpleJob *job : waitingJobs) {
         // ### for compatibility with the old scheduler we don't touch the running job, if any.
         // make sure that the job doesn't call back into Scheduler::cancelJob(); this would
@@ -311,7 +395,7 @@ bool ConnectedSlaveQueue::isIdle(Slave *slave)
     return it.value().runningJob == nullptr;
 }
 
-//private slot
+// private slot
 void ConnectedSlaveQueue::startRunnableJobs()
 {
     QSet<Slave *>::Iterator it = m_runnableSlaves.begin();
@@ -330,11 +414,11 @@ void ConnectedSlaveQueue::startRunnableJobs()
         jobs.runningJob = job;
 
         const QUrl url = job->url();
-        // no port is -1 in QUrl, but in kde3 we used 0 and the kioslaves assume that.
+        // no port is -1 in QUrl, but in kde3 we used 0 and the KIO workers assume that.
         const int port = url.port() == -1 ? 0 : url.port();
 
         if (slave->host() == QLatin1String("<reset>")) {
-            MetaData configData = SlaveConfig::self()->configData(url.scheme(), url.host());
+            MetaData configData = WorkerConfig::self()->configData(url.scheme(), url.host());
             slave->setConfig(configData);
             slave->setProtocol(url.scheme());
             slave->setHost(url.host(), port, url.userName(), url.password());
@@ -385,16 +469,16 @@ static void verifyRunningJobsCount(QHash<QString, HostQueue> *queues, int runnin
 #endif
 }
 
-ProtoQueue::ProtoQueue(int maxSlaves, int maxSlavesPerHost)
-    : m_maxConnectionsPerHost(maxSlavesPerHost ? maxSlavesPerHost : maxSlaves),
-      m_maxConnectionsTotal(qMax(maxSlaves, maxSlavesPerHost)),
-      m_runningJobsCount(0)
+ProtoQueue::ProtoQueue(int maxWorkers, int maxWorkersPerHost)
+    : m_maxConnectionsPerHost(maxWorkersPerHost ? maxWorkersPerHost : maxWorkers)
+    , m_maxConnectionsTotal(qMax(maxWorkers, maxWorkersPerHost))
+    , m_runningJobsCount(0)
 
 {
     /*qDebug() << "m_maxConnectionsTotal:" << m_maxConnectionsTotal
                  << "m_maxConnectionsPerHost:" << m_maxConnectionsPerHost;*/
     Q_ASSERT(m_maxConnectionsPerHost >= 1);
-    Q_ASSERT(maxSlaves >= maxSlavesPerHost);
+    Q_ASSERT(maxWorkers >= maxWorkersPerHost);
     m_startJobTimer.setSingleShot(true);
     connect(&m_startJobTimer, &QTimer::timeout, this, &ProtoQueue::startAJob);
 }
@@ -406,9 +490,8 @@ ProtoQueue::~ProtoQueue()
     // Clear the idle slaves in the keeper to avoid dangling pointers
     m_slaveKeeper.clear();
     for (Slave *slave : slaves) {
-        // kill the slave process, then remove the interface in our process
+        // kill the slave process and remove the interface in our process
         slave->kill();
-        slave->deref();
     }
 }
 
@@ -533,10 +616,9 @@ Slave *ProtoQueue::createSlave(const QString &protocol, SimpleJob *job, const QU
     QString errortext;
     Slave *slave = Slave::createSlave(protocol, url, error, errortext);
     if (slave) {
-        scheduler()->connect(slave, SIGNAL(slaveDied(KIO::Slave*)),
-                             SLOT(slotSlaveDied(KIO::Slave*)));
-        scheduler()->connect(slave, SIGNAL(slaveStatus(qint64,QByteArray,QString,bool)),
-                             SLOT(slotSlaveStatus(qint64,QByteArray,QString,bool)));
+        connect(slave, &Slave::slaveDied, scheduler(), [](KIO::Slave *slave) {
+            schedulerPrivate()->slotSlaveDied(slave);
+        });
     } else {
         qCWarning(KIO_CORE) << "couldn't create slave:" << errortext;
         if (job) {
@@ -565,25 +647,25 @@ QList<Slave *> ProtoQueue::allSlaves() const
     return ret;
 }
 
-//private slot
+// private slot
 void ProtoQueue::startAJob()
 {
     ensureNoDuplicates(&m_queuesBySerial);
     verifyRunningJobsCount(&m_queuesByHostname, m_runningJobsCount);
 
 #ifdef SCHEDULER_DEBUG
-    //qDebug() << "m_runningJobsCount:" << m_runningJobsCount;
+    // qDebug() << "m_runningJobsCount:" << m_runningJobsCount;
     auto it = m_queuesByHostname.cbegin();
     for (; it != m_queuesByHostname.cend(); ++it) {
         const QList<KIO::SimpleJob *> list = it.value().runningJobs();
         for (SimpleJob *job : list) {
-            //qDebug() << SimpleJobPrivate::get(job)->m_url;
+            // qDebug() << SimpleJobPrivate::get(job)->m_url;
         }
     }
 #endif
     if (m_runningJobsCount >= m_maxConnectionsTotal) {
 #ifdef SCHEDULER_DEBUG
-        //qDebug() << "not starting any jobs because maxConnectionsTotal has been reached.";
+        // qDebug() << "not starting any jobs because maxConnectionsTotal has been reached.";
 #endif
         return;
     }
@@ -626,7 +708,7 @@ void ProtoQueue::startAJob()
 
         if (slave) {
             jobPriv->m_slave = slave;
-            setupSlave(slave, jobPriv->m_url, jobPriv->m_protocol, jobPriv->m_proxyList, isNewSlave);
+            schedulerPrivate()->setupSlave(slave, jobPriv->m_url, jobPriv->m_protocol, jobPriv->m_proxyList, isNewSlave);
             startJob(startingJob, slave);
         } else {
             // dispose of our records about the job and mark the job as unknown
@@ -640,7 +722,7 @@ void ProtoQueue::startAJob()
         }
     } else {
 #ifdef SCHEDULER_DEBUG
-        //qDebug() << "not starting any jobs because there is no queued job.";
+        // qDebug() << "not starting any jobs because there is no queued job.";
 #endif
     }
 
@@ -649,147 +731,25 @@ void ProtoQueue::startAJob()
     }
 }
 
-class KIO::SchedulerPrivate
-{
-public:
-    SchedulerPrivate()
-        : q(new Scheduler()),
-          m_slaveOnHold(nullptr),
-          m_checkOnHold(true), // !! Always check with KLauncher for the first request
-          m_ignoreConfigReparse(false)
-    {
-    }
-
-    ~SchedulerPrivate()
-    {
-        delete q;
-        q = nullptr;
-        for (ProtoQueue *p : qAsConst(m_protocols)) {
-            const QList<KIO::Slave *> list = p->allSlaves();
-            for (Slave *slave : list) {
-                slave->kill();
-            }
-        }
-        qDeleteAll(m_protocols);
-    }
-
-    SchedulerPrivate(const SchedulerPrivate &) = delete;
-    SchedulerPrivate &operator=(const SchedulerPrivate &) = delete;
-
-    Scheduler *q;
-
-    Slave *m_slaveOnHold;
-    QUrl m_urlOnHold;
-    bool m_checkOnHold;
-    bool m_ignoreConfigReparse;
-
-    SessionData sessionData;
-
-    void doJob(SimpleJob *job);
-#if KIOCORE_BUILD_DEPRECATED_SINCE(4, 5)
-    void scheduleJob(SimpleJob *job);
-#endif
-    void setJobPriority(SimpleJob *job, int priority);
-    void cancelJob(SimpleJob *job);
-    void jobFinished(KIO::SimpleJob *job, KIO::Slave *slave);
-    void putSlaveOnHold(KIO::SimpleJob *job, const QUrl &url);
-    void removeSlaveOnHold();
-    Slave *getConnectedSlave(const QUrl &url, const KIO::MetaData &metaData);
-    bool assignJobToSlave(KIO::Slave *slave, KIO::SimpleJob *job);
-    bool disconnectSlave(KIO::Slave *slave);
-    void checkSlaveOnHold(bool b);
-    void publishSlaveOnHold();
-    Slave *heldSlaveForJob(KIO::SimpleJob *job);
-    bool isSlaveOnHoldFor(const QUrl &url);
-    void updateInternalMetaData(SimpleJob *job);
-
-    MetaData metaDataFor(const QString &protocol, const QStringList &proxyList, const QUrl &url);
-    void setupSlave(KIO::Slave *slave, const QUrl &url, const QString &protocol,
-                    const QStringList &proxyList, bool newSlave, const KIO::MetaData *config = nullptr);
-
-    void slotSlaveDied(KIO::Slave *slave);
-    void slotSlaveStatus(qint64 pid, const QByteArray &protocol,
-                         const QString &host, bool connected);
-
-    void slotReparseSlaveConfiguration(const QString &, const QDBusMessage &);
-    void slotSlaveOnHoldListChanged();
-
-    void slotSlaveConnected();
-    void slotSlaveError(int error, const QString &errorMsg);
-
-    ProtoQueue *protoQ(const QString &protocol, const QString &host)
-    {
-        ProtoQueue *pq = m_protocols.value(protocol, nullptr);
-        if (!pq) {
-            //qDebug() << "creating ProtoQueue instance for" << protocol;
-
-            const int maxSlaves = KProtocolInfo::maxSlaves(protocol);
-            int maxSlavesPerHost = -1;
-            if (!host.isEmpty()) {
-                bool ok = false;
-                const int value = SlaveConfig::self()->configData(protocol, host, QStringLiteral("MaxConnections")).toInt(&ok);
-                if (ok) {
-                    maxSlavesPerHost = value;
-                }
-            }
-            if (maxSlavesPerHost == -1) {
-                maxSlavesPerHost = KProtocolInfo::maxSlavesPerHost(protocol);
-            }
-            // Never allow maxSlavesPerHost to exceed maxSlaves.
-            pq = new ProtoQueue(maxSlaves, qMin(maxSlaves, maxSlavesPerHost));
-            m_protocols.insert(protocol, pq);
-        }
-        return pq;
-    }
-private:
-    QHash<QString, ProtoQueue *> m_protocols;
-};
-
-static QThreadStorage<SchedulerPrivate *> s_storage;
-static SchedulerPrivate *schedulerPrivate()
-{
-    if (!s_storage.hasLocalData()) {
-        s_storage.setLocalData(new SchedulerPrivate);
-    }
-    return s_storage.localData();
-}
-
-Scheduler *Scheduler::self()
-{
-    return schedulerPrivate()->q;
-}
-
-SchedulerPrivate *Scheduler::d_func()
-{
-    return schedulerPrivate();
-}
-
-//static
-Scheduler *scheduler()
-{
-    return schedulerPrivate()->q;
-}
-
-//static
-Slave *heldSlaveForJob(SimpleJob *job)
-{
-    return schedulerPrivate()->heldSlaveForJob(job);
-}
-
 Scheduler::Scheduler()
 {
     setObjectName(QStringLiteral("scheduler"));
 
+#ifndef KIO_ANDROID_STUB
     const QString dbusPath = QStringLiteral("/KIO/Scheduler");
     const QString dbusInterface = QStringLiteral("org.kde.KIO.Scheduler");
     QDBusConnection dbus = QDBusConnection::sessionBus();
     // Not needed, right? We just want to emit two signals.
-    //dbus.registerObject(dbusPath, this, QDBusConnection::ExportScriptableSlots |
+    // dbus.registerObject(dbusPath, this, QDBusConnection::ExportScriptableSlots |
     //                    QDBusConnection::ExportScriptableSignals);
-    dbus.connect(QString(), dbusPath, dbusInterface, QStringLiteral("reparseSlaveConfiguration"),
-                 this, SLOT(slotReparseSlaveConfiguration(QString,QDBusMessage)));
-    dbus.connect(QString(), dbusPath, dbusInterface, QStringLiteral("slaveOnHoldListChanged"),
-                 this, SLOT(slotSlaveOnHoldListChanged()));
+    dbus.connect(QString(),
+                 dbusPath,
+                 dbusInterface,
+                 QStringLiteral("reparseSlaveConfiguration"),
+                 this,
+                 SLOT(slotReparseSlaveConfiguration(QString, QDBusMessage)));
+    dbus.connect(QString(), dbusPath, dbusInterface, QStringLiteral("slaveOnHoldListChanged"), this, SLOT(slotSlaveOnHoldListChanged()));
+#endif
 }
 
 Scheduler::~Scheduler()
@@ -808,10 +768,18 @@ void Scheduler::scheduleJob(SimpleJob *job)
 }
 #endif
 
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 90)
 void Scheduler::setJobPriority(SimpleJob *job, int priority)
 {
     schedulerPrivate()->setJobPriority(job, priority);
 }
+#endif
+
+// static
+void Scheduler::setSimpleJobPriority(SimpleJob *job, int priority)
+{
+    schedulerPrivate()->setJobPriority(job, priority);
+}
 
 void Scheduler::cancelJob(SimpleJob *job)
 {
@@ -823,33 +791,55 @@ void Scheduler::jobFinished(KIO::SimpleJob *job, KIO::Slave *slave)
     schedulerPrivate()->jobFinished(job, slave);
 }
 
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 101)
 void Scheduler::putSlaveOnHold(KIO::SimpleJob *job, const QUrl &url)
 {
     schedulerPrivate()->putSlaveOnHold(job, url);
 }
+#endif
 
+void Scheduler::putWorkerOnHold(KIO::SimpleJob *job, const QUrl &url)
+{
+    schedulerPrivate()->putSlaveOnHold(job, url);
+}
+
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 101)
 void Scheduler::removeSlaveOnHold()
 {
     schedulerPrivate()->removeSlaveOnHold();
 }
+#endif
 
-void Scheduler::publishSlaveOnHold()
+void Scheduler::removeWorkerOnHold()
 {
-    schedulerPrivate()->publishSlaveOnHold();
+    schedulerPrivate()->removeSlaveOnHold();
 }
 
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 88)
+void Scheduler::publishSlaveOnHold()
+{
+}
+#endif
+
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 101)
 bool Scheduler::isSlaveOnHoldFor(const QUrl &url)
 {
     return schedulerPrivate()->isSlaveOnHoldFor(url);
 }
+#endif
+
+bool Scheduler::isWorkerOnHoldFor(const QUrl &url)
+{
+    return schedulerPrivate()->isSlaveOnHoldFor(url);
+}
 
 void Scheduler::updateInternalMetaData(SimpleJob *job)
 {
     schedulerPrivate()->updateInternalMetaData(job);
 }
 
-KIO::Slave *Scheduler::getConnectedSlave(const QUrl &url,
-        const KIO::MetaData &config)
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 91)
+KIO::Slave *Scheduler::getConnectedSlave(const QUrl &url, const KIO::MetaData &config)
 {
     return schedulerPrivate()->getConnectedSlave(url, config);
 }
@@ -863,62 +853,62 @@ bool Scheduler::disconnectSlave(KIO::Slave *slave)
 {
     return schedulerPrivate()->disconnectSlave(slave);
 }
+#endif
 
-bool Scheduler::connect(const char *signal, const QObject *receiver,
-                        const char *member)
+bool Scheduler::connect(const char *signal, const QObject *receiver, const char *member)
 {
     return QObject::connect(self(), signal, receiver, member);
 }
 
-bool Scheduler::connect(const QObject *sender, const char *signal,
-                        const QObject *receiver, const char *member)
+bool Scheduler::connect(const QObject *sender, const char *signal, const QObject *receiver, const char *member)
 {
     return QObject::connect(sender, signal, receiver, member);
 }
 
-bool Scheduler::disconnect(const QObject *sender, const char *signal,
-                           const QObject *receiver, const char *member)
+bool Scheduler::disconnect(const QObject *sender, const char *signal, const QObject *receiver, const char *member)
 {
     return QObject::disconnect(sender, signal, receiver, member);
 }
 
-bool Scheduler::connect(const QObject *sender, const char *signal,
-                        const char *member)
+bool Scheduler::connect(const QObject *sender, const char *signal, const char *member)
 {
     return QObject::connect(sender, signal, member);
 }
 
-void Scheduler::checkSlaveOnHold(bool b)
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 88)
+void Scheduler::checkSlaveOnHold(bool)
 {
-    schedulerPrivate()->checkSlaveOnHold(b);
 }
+#endif
 
 void Scheduler::emitReparseSlaveConfiguration()
 {
+#ifndef KIO_ANDROID_STUB
     // Do it immediately in this process, otherwise we might send a request before reparsing
     // (e.g. when changing useragent in the plugin)
     schedulerPrivate()->slotReparseSlaveConfiguration(QString(), QDBusMessage());
+#endif
 
     schedulerPrivate()->m_ignoreConfigReparse = true;
-    emit self()->reparseSlaveConfiguration(QString());
+    Q_EMIT self()->reparseSlaveConfiguration(QString());
 }
 
+#ifndef KIO_ANDROID_STUB
 void SchedulerPrivate::slotReparseSlaveConfiguration(const QString &proto, const QDBusMessage &)
 {
     if (m_ignoreConfigReparse) {
-        //qDebug() << "Ignoring signal sent by myself";
+        // qDebug() << "Ignoring signal sent by myself";
         m_ignoreConfigReparse = false;
         return;
     }
 
-    //qDebug() << "proto=" << proto;
+    // qDebug() << "proto=" << proto;
     KProtocolManager::reparseConfiguration();
-    SlaveConfig::self()->reset();
+    WorkerConfig::self()->reset();
     sessionData.reset();
     NetRC::self()->reload();
 
-    QHash<QString, ProtoQueue *>::ConstIterator it = proto.isEmpty() ? m_protocols.constBegin() :
-            m_protocols.constFind(proto);
+    QHash<QString, ProtoQueue *>::ConstIterator it = proto.isEmpty() ? m_protocols.constBegin() : m_protocols.constFind(proto);
     QHash<QString, ProtoQueue *>::ConstIterator endIt = m_protocols.constEnd();
 
     // not found?
@@ -939,40 +929,18 @@ void SchedulerPrivate::slotReparseSlaveConfiguration(const QString &proto, const
         }
     }
 }
+#endif
 
 void SchedulerPrivate::slotSlaveOnHoldListChanged()
 {
-    m_checkOnHold = true;
-}
-
-static bool mayReturnContent(int cmd, const QString &protocol)
-{
-    if (cmd == CMD_GET) {
-        return true;
-    }
-
-    if (cmd == CMD_MULTI_GET) {
-        return true;
-    }
-
-    if (cmd == CMD_SPECIAL && protocol.startsWith(QLatin1String("http"), Qt::CaseInsensitive)) {
-        return true;
-    }
-
-    return false;
 }
 
 void SchedulerPrivate::doJob(SimpleJob *job)
 {
-    //qDebug() << job;
+    // qDebug() << job;
     KIO::SimpleJobPrivate *const jobPriv = SimpleJobPrivate::get(job);
     jobPriv->m_proxyList.clear();
-    jobPriv->m_protocol = KProtocolManager::slaveProtocol(job->url(), jobPriv->m_proxyList);
-
-    if (mayReturnContent(jobCommand(job), jobPriv->m_protocol)) {
-        jobPriv->m_checkOnHold = m_checkOnHold;
-        m_checkOnHold = false;
-    }
+    jobPriv->m_protocol = KProtocolManager::workerProtocol(job->url(), jobPriv->m_proxyList);
 
     ProtoQueue *proto = protoQ(jobPriv->m_protocol, job->url().host());
     proto->queueJob(job);
@@ -981,41 +949,45 @@ void SchedulerPrivate::doJob(SimpleJob *job)
 #if KIOCORE_BUILD_DEPRECATED_SINCE(4, 5)
 void SchedulerPrivate::scheduleJob(SimpleJob *job)
 {
-    //qDebug() << job;
+    // qDebug() << job;
     setJobPriority(job, 1);
 }
 #endif
 
 void SchedulerPrivate::setJobPriority(SimpleJob *job, int priority)
 {
-    //qDebug() << job << priority;
+    // qDebug() << job << priority;
     const QString protocol = SimpleJobPrivate::get(job)->m_protocol;
     if (!protocol.isEmpty()) {
-        ProtoQueue *proto = protoQ(SimpleJobPrivate::get(job)->m_protocol, job->url().host());
+        ProtoQueue *proto = protoQ(protocol, job->url().host());
         proto->changeJobPriority(job, priority);
     }
 }
 
 void SchedulerPrivate::cancelJob(SimpleJob *job)
 {
+    KIO::SimpleJobPrivate *const jobPriv = SimpleJobPrivate::get(job);
     // this method is called all over the place in job.cpp, so just do this check here to avoid
     // much boilerplate in job code.
-    if (SimpleJobPrivate::get(job)->m_schedSerial == 0) {
-        //qDebug() << "Doing nothing because I don't know job" << job;
+    if (jobPriv->m_schedSerial == 0) {
+        // qDebug() << "Doing nothing because I don't know job" << job;
         return;
     }
     Slave *slave = jobSlave(job);
-    //qDebug() << job << slave;
-    if (slave) {
-        //qDebug() << "Scheduler: killing slave " << slave->slave_pid();
-        slave->kill();
-    }
+    // qDebug() << job << slave;
     jobFinished(job, slave);
+    if (slave) {
+        ProtoQueue *pq = m_protocols.value(jobPriv->m_protocol);
+        if (pq) {
+            pq->removeSlave(slave);
+        }
+        slave->kill(); // don't use slave after this!
+    }
 }
 
 void SchedulerPrivate::jobFinished(SimpleJob *job, Slave *slave)
 {
-    //qDebug() << job << slave;
+    // qDebug() << job << slave;
     KIO::SimpleJobPrivate *const jobPriv = SimpleJobPrivate::get(job);
 
     // make sure that we knew about the job!
@@ -1027,19 +999,18 @@ void SchedulerPrivate::jobFinished(SimpleJob *job, Slave *slave)
     }
 
     if (slave) {
-        // If we have internal meta-data, tell existing ioslaves to reload
+        // If we have internal meta-data, tell existing KIO workers to reload
         // their configuration.
         if (jobPriv->m_internalMetaData.count()) {
-            //qDebug() << "Updating ioslaves with new internal metadata information";
+            // qDebug() << "Updating KIO workers with new internal metadata information";
             ProtoQueue *queue = m_protocols.value(slave->protocol());
             if (queue) {
-                QListIterator<Slave *> it(queue->allSlaves());
-                while (it.hasNext()) {
-                    Slave *runningSlave = it.next();
+                const QList<Slave *> slaves = queue->allSlaves();
+                for (auto *runningSlave : slaves) {
                     if (slave->host() == runningSlave->host()) {
                         slave->setConfig(metaDataFor(slave->protocol(), jobPriv->m_proxyList, job->url()));
                         /*qDebug() << "Updated configuration of" << slave->protocol()
-                                     << "ioslave, pid=" << slave->slave_pid();*/
+                                     << "KIO worker, pid=" << slave->slave_pid();*/
                     }
                 }
             }
@@ -1054,17 +1025,10 @@ void SchedulerPrivate::jobFinished(SimpleJob *job, Slave *slave)
     jobPriv->m_internalMetaData.clear();
 }
 
-// static
-void setupSlave(KIO::Slave *slave, const QUrl &url, const QString &protocol,
-                const QStringList &proxyList, bool newSlave, const KIO::MetaData *config)
-{
-    schedulerPrivate()->setupSlave(slave, url, protocol, proxyList, newSlave, config);
-}
-
 MetaData SchedulerPrivate::metaDataFor(const QString &protocol, const QStringList &proxyList, const QUrl &url)
 {
     const QString host = url.host();
-    MetaData configData = SlaveConfig::self()->configData(protocol, host);
+    MetaData configData = WorkerConfig::self()->configData(protocol, host);
     sessionData.configDataFor(configData, protocol, host);
     if (proxyList.isEmpty()) {
         configData.remove(QStringLiteral("UseProxy"));
@@ -1074,8 +1038,8 @@ MetaData SchedulerPrivate::metaDataFor(const QString &protocol, const QStringLis
         configData[QStringLiteral("ProxyUrls")] = proxyList.join(QLatin1Char(','));
     }
 
-    if (configData.contains(QLatin1String("EnableAutoLogin")) &&
-            configData.value(QStringLiteral("EnableAutoLogin")).compare(QLatin1String("true"), Qt::CaseInsensitive) == 0) {
+    if (configData.contains(QLatin1String("EnableAutoLogin"))
+        && configData.value(QStringLiteral("EnableAutoLogin")).compare(QLatin1String("true"), Qt::CaseInsensitive) == 0) {
         NetRC::AutoLogin l;
         l.login = url.userName();
         bool usern = (protocol == QLatin1String("ftp"));
@@ -1096,20 +1060,22 @@ MetaData SchedulerPrivate::metaDataFor(const QString &protocol, const QStringLis
     return configData;
 }
 
-void SchedulerPrivate::setupSlave(KIO::Slave *slave, const QUrl &url, const QString &protocol,
-                                  const QStringList &proxyList, bool newSlave, const KIO::MetaData *config)
+void SchedulerPrivate::setupSlave(KIO::Slave *slave,
+                                  const QUrl &url,
+                                  const QString &protocol,
+                                  const QStringList &proxyList,
+                                  bool newSlave,
+                                  const KIO::MetaData *config)
 {
     int port = url.port();
-    if (port == -1) { // no port is -1 in QUrl, but in kde3 we used 0 and the kioslaves assume that.
+    if (port == -1) { // no port is -1 in QUrl, but in kde3 we used 0 and the KIO workers assume that.
         port = 0;
     }
     const QString host = url.host();
     const QString user = url.userName();
     const QString passwd = url.password();
 
-    if (newSlave || slave->host() != host || slave->port() != port ||
-            slave->user() != user || slave->passwd() != passwd) {
-
+    if (newSlave || slave->host() != host || slave->port() != port || slave->user() != user || slave->passwd() != passwd) {
         MetaData configData = metaDataFor(protocol, proxyList, url);
         if (config) {
             configData += *config;
@@ -1121,13 +1087,9 @@ void SchedulerPrivate::setupSlave(KIO::Slave *slave, const QUrl &url, const QStr
     }
 }
 
-void SchedulerPrivate::slotSlaveStatus(qint64, const QByteArray &, const QString &, bool)
-{
-}
-
 void SchedulerPrivate::slotSlaveDied(KIO::Slave *slave)
 {
-    //qDebug() << slave;
+    // qDebug() << slave;
     Q_ASSERT(slave);
     Q_ASSERT(!slave->isAlive());
     ProtoQueue *pq = m_protocols.value(slave->protocol());
@@ -1150,7 +1112,7 @@ void SchedulerPrivate::slotSlaveDied(KIO::Slave *slave)
 void SchedulerPrivate::putSlaveOnHold(KIO::SimpleJob *job, const QUrl &url)
 {
     Slave *slave = jobSlave(job);
-    //qDebug() << job << url << slave;
+    // qDebug() << job << url << slave;
     slave->disconnect(job);
     // prevent the fake death of the slave from trying to kill the job again;
     // cf. Slave::hold(const QUrl &url) called in SchedulerPrivate::publishSlaveOnHold().
@@ -1165,24 +1127,13 @@ void SchedulerPrivate::putSlaveOnHold(KIO::SimpleJob *job, const QUrl &url)
     m_slaveOnHold->suspend();
 }
 
-void SchedulerPrivate::publishSlaveOnHold()
-{
-    //qDebug() << m_slaveOnHold;
-    if (!m_slaveOnHold) {
-        return;
-    }
-
-    m_slaveOnHold->hold(m_urlOnHold);
-    emit q->slaveOnHoldListChanged();
-}
-
 bool SchedulerPrivate::isSlaveOnHoldFor(const QUrl &url)
 {
     if (url.isValid() && m_urlOnHold.isValid() && url == m_urlOnHold) {
         return true;
     }
 
-    return Slave::checkForHeldSlave(url);
+    return false;
 }
 
 Slave *SchedulerPrivate::heldSlaveForJob(SimpleJob *job)
@@ -1190,11 +1141,7 @@ Slave *SchedulerPrivate::heldSlaveForJob(SimpleJob *job)
     Slave *slave = nullptr;
     KIO::SimpleJobPrivate *const jobPriv = SimpleJobPrivate::get(job);
 
-    if (jobPriv->m_checkOnHold) {
-        slave = Slave::holdSlave(jobPriv->m_protocol, job->url());
-    }
-
-    if (!slave && m_slaveOnHold) {
+    if (m_slaveOnHold) {
         // Make sure that the job wants to do a GET or a POST, and with no offset
         const int cmd = jobPriv->m_command;
         bool canJobReuse = (cmd == CMD_GET || cmd == CMD_MULTI_GET);
@@ -1205,24 +1152,22 @@ Slave *SchedulerPrivate::heldSlaveForJob(SimpleJob *job)
                 KIO::MetaData outgoing = tJob->outgoingMetaData();
                 const QString resume = outgoing.value(QStringLiteral("resume"));
                 const QString rangeStart = outgoing.value(QStringLiteral("range-start"));
-                //qDebug() << "Resume metadata is" << resume;
+                // qDebug() << "Resume metadata is" << resume;
                 canJobReuse = (resume.isEmpty() || resume == QLatin1Char('0')) && (rangeStart.isEmpty() || rangeStart == QLatin1Char('0'));
             }
         }
 
         if (job->url() == m_urlOnHold) {
             if (canJobReuse) {
-                //qDebug() << "HOLD: Reusing held slave (" << m_slaveOnHold << ")";
+                // qDebug() << "HOLD: Reusing held slave (" << m_slaveOnHold << ")";
                 slave = m_slaveOnHold;
             } else {
-                //qDebug() << "HOLD: Discarding held slave (" << m_slaveOnHold << ")";
+                // qDebug() << "HOLD: Discarding held slave (" << m_slaveOnHold << ")";
                 m_slaveOnHold->kill();
             }
             m_slaveOnHold = nullptr;
             m_urlOnHold.clear();
         }
-    } else if (slave) {
-        //qDebug() << "HOLD: Reusing klauncher held slave (" << slave << ")";
     }
 
     return slave;
@@ -1230,7 +1175,7 @@ Slave *SchedulerPrivate::heldSlaveForJob(SimpleJob *job)
 
 void SchedulerPrivate::removeSlaveOnHold()
 {
-    //qDebug() << m_slaveOnHold;
+    // qDebug() << m_slaveOnHold;
     if (m_slaveOnHold) {
         m_slaveOnHold->kill();
     }
@@ -1238,51 +1183,77 @@ void SchedulerPrivate::removeSlaveOnHold()
     m_urlOnHold.clear();
 }
 
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 91)
 Slave *SchedulerPrivate::getConnectedSlave(const QUrl &url, const KIO::MetaData &config)
 {
     QStringList proxyList;
-    const QString protocol = KProtocolManager::slaveProtocol(url, proxyList);
+    const QString protocol = KProtocolManager::workerProtocol(url, proxyList);
     ProtoQueue *pq = protoQ(protocol, url.host());
 
-    Slave *slave = pq->createSlave(protocol, /* job */nullptr, url);
+    Slave *slave = pq->createSlave(protocol, /* job */ nullptr, url);
     if (slave) {
         setupSlave(slave, url, protocol, proxyList, true, &config);
         pq->m_connectedSlaveQueue.addSlave(slave);
 
         slave->send(CMD_CONNECT);
-        q->connect(slave, SIGNAL(connected()),
-                   SLOT(slotSlaveConnected()));
-        q->connect(slave, SIGNAL(error(int,QString)),
-                   SLOT(slotSlaveError(int,QString)));
+        q->connect(slave, SIGNAL(connected()), SLOT(slotSlaveConnected()));
+        q->connect(slave, SIGNAL(error(int, QString)), SLOT(slotSlaveError(int, QString)));
     }
-    //qDebug() << url << slave;
+    // qDebug() << url << slave;
     return slave;
 }
 
 void SchedulerPrivate::slotSlaveConnected()
 {
-    //qDebug();
+    // qDebug();
     Slave *slave = static_cast<Slave *>(q->sender());
     slave->setConnected(true);
     q->disconnect(slave, SIGNAL(connected()), q, SLOT(slotSlaveConnected()));
-    emit q->slaveConnected(slave);
+    Q_EMIT q->slaveConnected(slave);
 }
 
 void SchedulerPrivate::slotSlaveError(int errorNr, const QString &errorMsg)
 {
     Slave *slave = static_cast<Slave *>(q->sender());
-    //qDebug() << slave << errorNr << errorMsg;
+    // qDebug() << slave << errorNr << errorMsg;
     ProtoQueue *pq = protoQ(slave->protocol(), slave->host());
     if (!slave->isConnected() || pq->m_connectedSlaveQueue.isIdle(slave)) {
         // Only forward to application if slave is idle or still connecting.
         // ### KDE5: can we remove this apparently arbitrary behavior and just always emit SlaveError?
-        emit q->slaveError(slave, errorNr, errorMsg);
+        Q_EMIT q->slaveError(slave, errorNr, errorMsg);
     }
 }
+#endif
 
+ProtoQueue *SchedulerPrivate::protoQ(const QString &protocol, const QString &host)
+{
+    ProtoQueue *pq = m_protocols.value(protocol, nullptr);
+    if (!pq) {
+        // qDebug() << "creating ProtoQueue instance for" << protocol;
+
+        const int maxWorkers = KProtocolInfo::maxWorkers(protocol);
+        int maxWorkersPerHost = -1;
+        if (!host.isEmpty()) {
+            bool ok = false;
+            const int value = WorkerConfig::self()->configData(protocol, host, QStringLiteral("MaxConnections")).toInt(&ok);
+            if (ok) {
+                maxWorkersPerHost = value;
+            }
+        }
+        if (maxWorkersPerHost == -1) {
+            maxWorkersPerHost = KProtocolInfo::maxWorkersPerHost(protocol);
+        }
+        // Never allow maxWorkersPerHost to exceed maxWorkers.
+        pq = new ProtoQueue(maxWorkers, qMin(maxWorkers, maxWorkersPerHost));
+        m_protocols.insert(protocol, pq);
+    }
+    return pq;
+}
+
+#if KIOCORE_BUILD_DEPRECATED_SINCE(5, 91)
 bool SchedulerPrivate::assignJobToSlave(KIO::Slave *slave, SimpleJob *job)
 {
-    //qDebug() << slave << job;
+    // qDebug() << slave << job;
     // KDE5: queueing of jobs can probably be removed, it provides very little benefit
     ProtoQueue *pq = m_protocols.value(slave->protocol());
     if (pq) {
@@ -1294,31 +1265,29 @@ bool SchedulerPrivate::assignJobToSlave(KIO::Slave *slave, SimpleJob *job)
 
 bool SchedulerPrivate::disconnectSlave(KIO::Slave *slave)
 {
-    //qDebug() << slave;
+    // qDebug() << slave;
     ProtoQueue *pq = m_protocols.value(slave->protocol());
     return (pq ? pq->m_connectedSlaveQueue.removeSlave(slave) : false);
 }
-
-void SchedulerPrivate::checkSlaveOnHold(bool b)
-{
-    //qDebug() << b;
-    m_checkOnHold = b;
-}
+#endif
 
 void SchedulerPrivate::updateInternalMetaData(SimpleJob *job)
 {
     KIO::SimpleJobPrivate *const jobPriv = SimpleJobPrivate::get(job);
     // Preserve all internal meta-data so they can be sent back to the
-    // ioslaves as needed...
+    // KIO workers as needed...
     const QUrl jobUrl = job->url();
-    //qDebug() << job << jobPriv->m_internalMetaData;
+
+    const QLatin1String currHostToken("{internal~currenthost}");
+    const QLatin1String allHostsToken("{internal~allhosts}");
+    // qDebug() << job << jobPriv->m_internalMetaData;
     QMapIterator<QString, QString> it(jobPriv->m_internalMetaData);
     while (it.hasNext()) {
         it.next();
-        if (it.key().startsWith(QLatin1String("{internal~currenthost}"), Qt::CaseInsensitive)) {
-            SlaveConfig::self()->setConfigData(jobUrl.scheme(), jobUrl.host(), it.key().mid(22), it.value());
-        } else if (it.key().startsWith(QLatin1String("{internal~allhosts}"), Qt::CaseInsensitive)) {
-            SlaveConfig::self()->setConfigData(jobUrl.scheme(), QString(), it.key().mid(19), it.value());
+        if (it.key().startsWith(currHostToken, Qt::CaseInsensitive)) {
+            WorkerConfig::self()->setConfigData(jobUrl.scheme(), jobUrl.host(), it.key().mid(currHostToken.size()), it.value());
+        } else if (it.key().startsWith(allHostsToken, Qt::CaseInsensitive)) {
+            WorkerConfig::self()->setConfigData(jobUrl.scheme(), QString(), it.key().mid(allHostsToken.size()), it.value());
         }
     }
 }
