@@ -25,22 +25,19 @@
 #include <QTimeLine>
 #include <QTimer>
 #include <QToolTip>
+#include <QVariantAnimation>
+#include <kio/deleteortrashjob.h>
 
 #include <KColorScheme>
 #include <KColorUtils>
 #include <KConfig>
 #include <KConfigGroup>
-#include <KIO/JobUiDelegateFactory>
 #include <KJob>
-#include <KJobWidgets>
 #include <KLocalizedString>
-#include <KMessageBox>
 #include <KSharedConfig>
 #include <defaults-kfile.h> // ConfigGroup, PlacesIconsAutoresize, PlacesIconsStaticSize
 #include <kdirnotify.h>
-#include <kio/emptytrashjob.h>
 #include <kio/filesystemfreespacejob.h>
-#include <kio/jobuidelegate.h>
 #include <kmountpoint.h>
 #include <kpropertiesdialog.h>
 #include <solid/opticaldisc.h>
@@ -48,7 +45,6 @@
 #include <solid/storageaccess.h>
 #include <solid/storagedrive.h>
 #include <solid/storagevolume.h>
-#include <widgetsaskuseractionhandler.h>
 
 #include <chrono>
 #include <cmath>
@@ -152,8 +148,13 @@ void KFilePlacesViewDelegate::paint(QPainter *painter, const QStyleOptionViewIte
 
     QApplication::style()->drawPrimitive(QStyle::PE_PanelItemViewItem, &opt, painter);
 
+    const auto accessibility = placesModel->deviceAccessibility(index);
+    const bool isBusy = (accessibility == KFilePlacesModel::SetupInProgress || accessibility == KFilePlacesModel::TeardownInProgress);
+
     QIcon actionIcon;
-    if (placesModel->isTeardownOverlayRecommended(index)) {
+    if (isBusy) {
+        actionIcon = QIcon::fromTheme(QStringLiteral("view-refresh"));
+    } else if (placesModel->isTeardownOverlayRecommended(index)) {
         actionIcon = QIcon::fromTheme(QStringLiteral("media-eject"));
     }
 
@@ -180,16 +181,32 @@ void KFilePlacesViewDelegate::paint(QPainter *painter, const QStyleOptionViewIte
     painter->drawPixmap(point, pm);
 
     if (!actionIcon.isNull()) {
-        QPoint actionPos(isLTR ? opt.rect.right() - actionAreaWidth : opt.rect.left() + s_lateralMargin,
-                         opt.rect.top() + (opt.rect.height() - actionIconSize()) / 2);
-        QIcon::Mode actionMode = QIcon::Normal;
+        const int iconSize = actionIconSize();
+        QIcon::Mode mode = QIcon::Normal;
         if (selectedAndActive) {
-            actionMode = QIcon::Selected;
+            mode = QIcon::Selected;
         } else if (m_hoveredAction == index) {
-            actionMode = QIcon::Active;
+            mode = QIcon::Active;
+        }
+
+        const QPixmap pixmap = actionIcon.pixmap(iconSize, iconSize, mode);
+
+        const QRectF rect(isLTR ? opt.rect.right() - actionAreaWidth : opt.rect.left() + s_lateralMargin,
+                          opt.rect.top() + (opt.rect.height() - iconSize) / 2,
+                          iconSize,
+                          iconSize);
+
+        if (isBusy) {
+            painter->save();
+            painter->setRenderHint(QPainter::SmoothPixmapTransform);
+            painter->translate(rect.center());
+            painter->rotate(m_busyAnimationRotation);
+            painter->translate(QPointF(-rect.width() / 2.0, -rect.height() / 2.0));
+            painter->drawPixmap(0, 0, pixmap);
+            painter->restore();
+        } else {
+            painter->drawPixmap(rect.topLeft(), pixmap);
         }
-        QPixmap actionPix = actionIcon.pixmap(actionIconSize(), actionIconSize(), actionMode);
-        painter->drawPixmap(actionPos, actionPix);
     }
 
     if (changePalette) {
@@ -324,6 +341,11 @@ void KFilePlacesViewDelegate::setAppearingItemProgress(qreal value)
     }
 }
 
+void KFilePlacesViewDelegate::setDeviceBusyAnimationRotation(qreal angle)
+{
+    m_busyAnimationRotation = angle;
+}
+
 void KFilePlacesViewDelegate::addDisappearingItem(const QModelIndex &index)
 {
     m_disappearingItems << index;
@@ -688,6 +710,8 @@ public:
     void enableSmoothItemResizing();
     void slotEmptyTrash();
 
+    void deviceBusyAnimationValueChanged(const QVariant &value);
+
     KFilePlacesView *const q;
 
     KFilePlacesEventWatcher *const m_watcher;
@@ -706,12 +730,13 @@ public:
 
     KFilePlacesView::TeardownFunction m_teardownFunction = nullptr;
 
-    std::unique_ptr<KIO::WidgetsAskUserActionHandler> m_askUserHandler;
-
     QTimeLine m_adaptItemsTimeline;
     QTimeLine m_itemAppearTimeline;
     QTimeLine m_itemDisappearTimeline;
 
+    QVariantAnimation m_deviceBusyAnimation;
+    QVector<QPersistentModelIndex> m_busyDevices;
+
     QRect m_dropRect;
     QPersistentModelIndex m_dropIndex;
 
@@ -813,6 +838,15 @@ KFilePlacesView::KFilePlacesView(QWidget *parent)
     d->m_itemDisappearTimeline.setUpdateInterval(5);
     d->m_itemDisappearTimeline.setEasingCurve(QEasingCurve::InOutSine);
 
+    // Adapted from KBusyIndicatorWidget
+    d->m_deviceBusyAnimation.setLoopCount(-1);
+    d->m_deviceBusyAnimation.setDuration(2000);
+    d->m_deviceBusyAnimation.setStartValue(0);
+    d->m_deviceBusyAnimation.setEndValue(360);
+    connect(&d->m_deviceBusyAnimation, &QVariantAnimation::valueChanged, this, [this](const QVariant &value) {
+        d->deviceBusyAnimationValueChanged(value);
+    });
+
     viewport()->installEventFilter(d->m_watcher);
     connect(d->m_watcher, &KFilePlacesEventWatcher::entryMiddleClicked, this, [this](const QModelIndex &index) {
         if (qGuiApp->keyboardModifiers() == Qt::ShiftModifier && isSignalConnected(QMetaMethod::fromSignal(&KFilePlacesView::activeTabRequested))) {
@@ -1031,24 +1065,13 @@ void KFilePlacesViewPrivate::writeConfig()
 void KFilePlacesViewPrivate::slotEmptyTrash()
 {
     auto *parentWindow = q->window();
+
     using AskIface = KIO::AskUserActionInterface;
-
-    if (!m_askUserHandler) {
-        m_askUserHandler.reset(new KIO::WidgetsAskUserActionHandler{});
-
-        auto slotAskResult = [parentWindow](bool allowDelete, const QList<QUrl> &, AskIface::DeletionType, QWidget *parent) {
-            if (parent != parentWindow || !allowDelete) {
-                return;
-            }
-
-            KIO::Job *job = KIO::emptyTrash();
-            job->setUiDelegate(KIO::createDefaultJobUiDelegate(KJobUiDelegate::AutoHandlingEnabled, parentWindow));
-        };
-
-        q->connect(m_askUserHandler.get(), &AskIface::askUserDeleteResult, q, slotAskResult);
-    }
-
-    m_askUserHandler->askUserDelete(QList<QUrl>{}, AskIface::EmptyTrash, AskIface::DefaultConfirmation, parentWindow);
+    auto *emptyTrashJob = new KIO::DeleteOrTrashJob(QList<QUrl>{}, //
+                                                    AskIface::EmptyTrash,
+                                                    AskIface::DefaultConfirmation,
+                                                    parentWindow);
+    emptyTrashJob->start();
 }
 
 void KFilePlacesView::contextMenuEvent(QContextMenuEvent *event)
@@ -1103,7 +1126,9 @@ void KFilePlacesView::contextMenuEvent(QContextMenuEvent *event)
             teardown = placesModel->teardownActionForIndex(index);
             if (teardown) {
                 teardown->setParent(&menu);
-                teardown->setEnabled(placesModel->isTeardownAllowed(index));
+                if (!placesModel->isTeardownAllowed(index)) {
+                    teardown->setEnabled(false);
+                }
             }
 
             if (placesModel->setupNeeded(index)) {
@@ -1994,10 +2019,39 @@ void KFilePlacesViewPrivate::enableSmoothItemResizing()
     m_smoothItemResizing = true;
 }
 
+void KFilePlacesViewPrivate::deviceBusyAnimationValueChanged(const QVariant &value)
+{
+    m_delegate->setDeviceBusyAnimationRotation(value.toReal());
+    for (const auto &idx : std::as_const(m_busyDevices)) {
+        q->update(idx);
+    }
+}
+
 void KFilePlacesView::dataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight, const QVector<int> &roles)
 {
     QListView::dataChanged(topLeft, bottomRight, roles);
     d->adaptItemSize();
+
+    if ((roles.isEmpty() || roles.contains(KFilePlacesModel::DeviceAccessibilityRole)) && d->shouldAnimate()) {
+        QVector<QPersistentModelIndex> busyDevices;
+
+        auto *placesModel = qobject_cast<KFilePlacesModel *>(model());
+        for (int i = 0; i < placesModel->rowCount(); ++i) {
+            const QModelIndex idx = placesModel->index(i, 0);
+            const auto accessibility = placesModel->deviceAccessibility(idx);
+            if (accessibility == KFilePlacesModel::SetupInProgress || accessibility == KFilePlacesModel::TeardownInProgress) {
+                busyDevices.append(QPersistentModelIndex(idx));
+            }
+        }
+
+        d->m_busyDevices = busyDevices;
+
+        if (busyDevices.isEmpty()) {
+            d->m_deviceBusyAnimation.stop();
+        } else {
+            d->m_deviceBusyAnimation.start();
+        }
+    }
 }
 
 #include "moc_kfileplacesview_p.cpp"
