@@ -10,13 +10,13 @@
 #include "kurlnavigator.h"
 #include "kcoreurlnavigator.h"
 
+#include "../utils_p.h"
 #include "kurlnavigatorbutton_p.h"
 #include "kurlnavigatordropdownbutton_p.h"
 #include "kurlnavigatorpathselectoreventfilter_p.h"
 #include "kurlnavigatorplacesselector_p.h"
 #include "kurlnavigatorprotocolcombo_p.h"
 #include "kurlnavigatortogglebutton_p.h"
-#include "urlutil_p.h"
 
 #include <KIO/StatJob>
 #include <KLocalizedString>
@@ -71,6 +71,10 @@ public:
 
     /** Applies the edited URL in m_pathBox to the URL navigator */
     void applyUncommittedUrl();
+    void slotApplyUrl(QUrl url);
+    // Returns true if "text" matched a URI filter (i.e. was fitlered),
+    // otherwise returns false
+    bool slotCheckFilters(const QString &text);
 
     void slotReturnPressed();
     void slotProtocolChanged(const QString &);
@@ -159,8 +163,6 @@ public:
      */
     QUrl retrievePlaceUrl() const;
 
-    void removeTrailingSlash(QString &url) const;
-
     KUrlNavigator *const q;
 
     QHBoxLayout *m_layout = new QHBoxLayout(q);
@@ -289,61 +291,86 @@ void KUrlNavigatorPrivate::appendWidget(QWidget *widget, int stretch)
     m_layout->insertWidget(m_layout->count() - 1, widget, stretch);
 }
 
-void KUrlNavigatorPrivate::applyUncommittedUrl()
+void KUrlNavigatorPrivate::slotApplyUrl(QUrl url)
 {
-    auto applyUrl = [this](QUrl url) {
-        // Parts of the following code have been taken from the class KateFileSelector
-        // located in kate/app/katefileselector.hpp of Kate.
-        // SPDX-FileCopyrightText: 2001 Christoph Cullmann <cullmann@kde.org>
-        // SPDX-FileCopyrightText: 2001 Joseph Wenninger <jowenn@kde.org>
-        // SPDX-FileCopyrightText: 2001 Anders Lund <anders.lund@lund.tdcadsl.dk>
+    // Parts of the following code have been taken from the class KateFileSelector
+    // located in kate/app/katefileselector.hpp of Kate.
+    // SPDX-FileCopyrightText: 2001 Christoph Cullmann <cullmann@kde.org>
+    // SPDX-FileCopyrightText: 2001 Joseph Wenninger <jowenn@kde.org>
+    // SPDX-FileCopyrightText: 2001 Anders Lund <anders.lund@lund.tdcadsl.dk>
 
-        // For example "desktop:/" _not_ "desktop:", see the comment in slotProtocolChanged()
-        if (!url.isEmpty() && url.path().isEmpty() && KProtocolInfo::protocolClass(url.scheme()) == QLatin1String(":local")) {
-            url.setPath(QStringLiteral("/"));
-        }
+    // For example "desktop:/" _not_ "desktop:", see the comment in slotProtocolChanged()
+    if (!url.isEmpty() && url.path().isEmpty() && KProtocolInfo::protocolClass(url.scheme()) == QLatin1String(":local")) {
+        url.setPath(QStringLiteral("/"));
+    }
 
-        const auto urlStr = url.toString();
-        QStringList urls = m_pathBox->urls();
-        urls.removeAll(urlStr);
-        urls.prepend(urlStr);
-        m_pathBox->setUrls(urls, KUrlComboBox::RemoveBottom);
+    const auto urlStr = url.toString();
+    QStringList urls = m_pathBox->urls();
+    urls.removeAll(urlStr);
+    urls.prepend(urlStr);
+    m_pathBox->setUrls(urls, KUrlComboBox::RemoveBottom);
 
-        q->setLocationUrl(url);
-        // The URL might have been adjusted by KUrlNavigator::setUrl(), hence
-        // synchronize the result in the path box.
-        m_pathBox->setUrl(q->locationUrl());
-    };
-
-    const QString text = m_pathBox->currentText().trimmed();
+    q->setLocationUrl(url);
+    // The URL might have been adjusted by KUrlNavigator::setUrl(), hence
+    // synchronize the result in the path box.
+    m_pathBox->setUrl(q->locationUrl());
+}
 
+bool KUrlNavigatorPrivate::slotCheckFilters(const QString &text)
+{
     KUriFilterData filteredData(text);
     filteredData.setCheckForExecutables(false);
     // Using kshorturifilter to fix up e.g. "ftp.kde.org" ---> "ftp://ftp.kde.org"
     const auto filtersList = QStringList{QStringLiteral("kshorturifilter")};
-    if (KUriFilter::self()->filterUri(filteredData, filtersList)) {
-        applyUrl(filteredData.uri()); // The text was filtered
-        return;
+    const bool wasFiltered = KUriFilter::self()->filterUri(filteredData, filtersList);
+    if (wasFiltered) {
+        slotApplyUrl(filteredData.uri()); // The text was filtered
+    }
+    return wasFiltered;
+}
+
+void KUrlNavigatorPrivate::applyUncommittedUrl()
+{
+    const QString text = m_pathBox->currentText().trimmed();
+    QUrl url = q->locationUrl();
+
+    // Using the stat job below, check if the url and text match a local dir; but first
+    // handle a special case where "url" is empty in the unittests which use
+    // KUrlNavigator::setLocationUrl(QUrl()); in practice (e.g. in Dolphin, or KFileWidget),
+    // locationUrl() is never empty
+    if (url.isEmpty() && !text.isEmpty()) {
+        if (slotCheckFilters(text)) {
+            return;
+        }
     }
 
-    QUrl url = q->locationUrl();
-    QString path = url.path();
-    if (!path.endsWith(QLatin1Char('/'))) {
-        path += QLatin1Char('/');
+    // Treat absolute paths as absolute paths.
+    // Relative paths get appended to the current path.
+    if (text.startsWith(QLatin1Char('/'))) {
+        url.setPath(text);
+    } else {
+        url.setPath(Utils::concatPaths(url.path(), text));
     }
-    url.setPath(path + text);
 
     // Dirs and symlinks to dirs
     constexpr auto details = KIO::StatBasic | KIO::StatResolveSymlink;
     auto *job = KIO::statDetails(url, KIO::StatJob::DestinationSide, details, KIO::HideProgressInfo);
-    q->connect(job, &KJob::result, q, [job, text, applyUrl]() {
-        // If there is a dir matching "text" relative to the current url, use that, e.g.
-        // typing "bar" while at "/path/to/foo", the url becomes "/path/to/foo/bar/"
+    q->connect(job, &KJob::result, q, [this, job, text]() {
+        // If there is a dir matching "text" relative to the current url, use that, e.g.:
+        // - typing "bar" while at "/path/to/foo" ---> "/path/to/foo/bar/"
+        // - typing ".config" while at "/home/foo" ---> "/home/foo/.config"
         if (!job->error() && job->statResult().isDir()) {
-            applyUrl(job->url());
-        } else { // ... otherwise fallback to whatever QUrl::fromUserInput() returns
-            applyUrl(QUrl::fromUserInput(text));
+            slotApplyUrl(job->url());
+            return;
         }
+
+        // Check if text matches a URI filter
+        if (slotCheckFilters(text)) {
+            return;
+        }
+
+        // ... otherwise fallback to whatever QUrl::fromUserInput() returns
+        slotApplyUrl(QUrl::fromUserInput(text));
     });
 }
 
@@ -357,7 +384,10 @@ void KUrlNavigatorPrivate::slotReturnPressed()
         // Pressing Ctrl+Return automatically switches back to the breadcrumb mode.
         // The switch must be done asynchronously, as we are in the context of the
         // editor.
-        QMetaObject::invokeMethod(q, "switchToBreadcrumbMode", Qt::QueuedConnection);
+        auto switchModeFunc = [this]() {
+            switchToBreadcrumbMode();
+        };
+        QMetaObject::invokeMethod(q, switchModeFunc, Qt::QueuedConnection);
     }
 }
 
@@ -622,8 +652,7 @@ void KUrlNavigatorPrivate::updateContent()
         if (!placeUrl.isValid()) {
             placeUrl = retrievePlaceUrl();
         }
-        QString placePath = placeUrl.path();
-        removeTrailingSlash(placePath);
+        QString placePath = Utils::trailingSlashRemoved(placeUrl.path());
 
         const int startIndex = placePath.count(QLatin1Char('/'));
         updateButtons(startIndex);
@@ -732,11 +761,11 @@ void KUrlNavigatorPrivate::updateButtonVisibility()
         availableWidth -= m_protocols->width();
     }
 
-    auto MinWidth = [](const KUrlNavigatorButton *button) {
-        return button->minimumWidth();
-    };
     // Check whether buttons must be hidden at all...
-    const int requiredButtonWidth = std::transform_reduce(m_navButtons.cbegin(), m_navButtons.cend(), 0, std::plus<>(), MinWidth);
+    int requiredButtonWidth = 0;
+    for (const auto *button : std::as_const(m_navButtons)) {
+        requiredButtonWidth += button->minimumWidth();
+    }
 
     if (requiredButtonWidth > availableWidth) {
         // At least one button must be hidden. This implies that the
@@ -873,14 +902,6 @@ QUrl KUrlNavigatorPrivate::retrievePlaceUrl() const
     return currentUrl;
 }
 
-void KUrlNavigatorPrivate::removeTrailingSlash(QString &url) const
-{
-    const int length = url.length();
-    if ((length > 0) && (url.at(length - 1) == QLatin1Char('/'))) {
-        url.remove(length - 1, 1);
-    }
-}
-
 // ------------------------------------------------------------------------------------------------
 
 KUrlNavigator::KUrlNavigator(QWidget *parent)
