@@ -10,14 +10,17 @@
 #include "accessmanager.h"
 
 #include "accessmanagerreply_p.h"
-#include "job.h"
 #include "kio_widgets_debug.h"
 #include "scheduler.h"
+#include <kio/mimetypejob.h>
+#include <kio/statjob.h>
+#include <kio/storedtransferjob.h>
+#include <kprotocolinfo.h>
+
 #include <KConfigGroup>
 #include <KJobWidgets>
 #include <KLocalizedString>
 #include <KSharedConfig>
-#include <kprotocolinfo.h>
 
 #include <QDBusInterface>
 #include <QDBusReply>
@@ -30,7 +33,6 @@
 #include <QUrl>
 #include <QWidget>
 
-
 static const QNetworkRequest::Attribute gSynchronousNetworkRequestAttribute = QNetworkRequest::SynchronousRequestAttribute;
 
 static qint64 sizeFromRequest(const QNetworkRequest &req)
@@ -96,10 +98,7 @@ AccessManager::AccessManager(QObject *parent)
     setCookieJar(new KIO::Integration::CookieJar);
 }
 
-AccessManager::~AccessManager()
-{
-    delete d;
-}
+AccessManager::~AccessManager() = default;
 
 void AccessManager::setExternalContentAllowed(bool allowed)
 {
@@ -234,7 +233,7 @@ QNetworkReply *AccessManager::createRequest(Operation op, const QNetworkRequest
 
             QVariant len = req.header(QNetworkRequest::ContentLengthHeader);
             if (len.isValid()) {
-                storedJob->setTotalSize(len.toInt());
+                storedJob->setTotalSize(len.toULongLong());
             }
 
             kioJob = storedJob;
@@ -287,10 +286,10 @@ QNetworkReply *AccessManager::createRequest(Operation op, const QNetworkRequest
     // Set the job priority
     switch (req.priority()) {
     case QNetworkRequest::HighPriority:
-        KIO::Scheduler::setJobPriority(kioJob, -5);
+        KIO::Scheduler::setSimpleJobPriority(kioJob, -5);
         break;
     case QNetworkRequest::LowPriority:
-        KIO::Scheduler::setJobPriority(kioJob, 5);
+        KIO::Scheduler::setSimpleJobPriority(kioJob, 5);
         break;
     default:
         break;
@@ -337,10 +336,10 @@ QNetworkReply *AccessManager::createRequest(Operation op, const QNetworkRequest
         // Set the job priority
         switch (req.priority()) {
         case QNetworkRequest::HighPriority:
-            KIO::Scheduler::setJobPriority(kioJob, -5);
+            KIO::Scheduler::setSimpleJobPriority(kioJob, -5);
             break;
         case QNetworkRequest::LowPriority:
-            KIO::Scheduler::setJobPriority(kioJob, 5);
+            KIO::Scheduler::setSimpleJobPriority(kioJob, 5);
             break;
         default:
             break;
@@ -471,10 +470,7 @@ CookieJar::CookieJar(QObject *parent)
     reparseConfiguration();
 }
 
-CookieJar::~CookieJar()
-{
-    delete d;
-}
+CookieJar::~CookieJar() = default;
 
 WId CookieJar::windowId() const
 {
